# -*- coding: utf-8 -*-

"""
Building a workflow in Simmate can sometimes involve using custom database
tables, so before you start editting this file, make sure you have gone through
and editted the `models.py` file.

Below is an example of a simple VaspWorkflow, which is used to run a single VASP
calculation.

As your need grow, you may need more complex Workflows. For complex workflows that
require several calculations, other python logic, and parallelization, make
sure you read through the documentation for 
[the Workflow class](https://jacksund.github.io/simmate/simmate/workflow_engine/workflow.html) as well as relevent subclasses like the
[S3Workflow](https://jacksund.github.io/simmate/simmate/workflow_engine/s3_workflow.html)
or 
[VaspWorkflow](https://jacksund.github.io/simmate/simmate/calculators/vasp/tasks/base.html)
"""

from simmate.workflow_engine import Workflow
from simmate.calculators.vasp.inputs import Incar
from simmate.calculators.vasp.workflows.base import VaspWorkflow
from simmate.calculators.vasp.inputs.potcar_mappings import (
    PBE_ELEMENT_MAPPINGS_LOW_QUALITY,
)
from simmate.calculators.vasp.error_handlers import (
    Unconverged,
    NonConverging,
    Frozen,
    Walltime,
    # There are many more available too!
)

# Import our tables from the other files.
from .models import MyCustomTable1, MyCustomTable2

# -----------------------------------------------------------------------------
# Our first example shows off the basics
# -----------------------------------------------------------------------------


class Example__Python__MyExample1(Workflow):

    database_table = MyCustomTable1

    @staticmethod
    def run_config(x, y):
        # This is a boring workflow because it just saves the input values
        # to our table. You can get much more creative with workflows though
        new_entry = MyCustomTable2(
            custom_column_01=x,
            custom_column_02=y,
        )
        new_entry.save()

        return new_entry.id


# -----------------------------------------------------------------------------
# The example below will be especially helpful for VASP users
# -----------------------------------------------------------------------------


class Relaxation__Vasp__MyExample2(VaspWorkflow):

    # This uses the PBE functional with POTCARs that have lower electron counts
    # and convergence criteria when available.
    functional = "PBE"
    potcar_mappings = PBE_ELEMENT_MAPPINGS_LOW_QUALITY

    # These are all input settings for this task.
    incar = dict(
        # These settings are the same for all structures regardless of composition.
        PREC="Normal",
        EDIFF=1e-5,
        ENCUT=450,  # !!! Should this be based on the element type?
        ISIF=3,
        NSW=100,
        IBRION=1,
        POTIM=0.02,
        LCHARG=False,
        LWAVE=False,
        KSPACING=0.4,
        # The type of smearing we use depends on if we have a metal, semiconductor,
        # or insulator. So we need to decide this using a keyword modifier.
        multiple_keywords__smart_ismear={
            "metal": dict(
                ISMEAR=1,
                SIGMA=0.06,
            ),
            "non-metal": dict(
                ISMEAR=0,
                SIGMA=0.05,
            ),
        },
        # This is a dummy example of using custom modifiers. This would set the
        # incar value of EXAMPLE=16 for structure with 2 sites (2*8=16). Note,
        # we registered "multiply_nsites" logic below.
        EXAMPLE__multiply_nsites=8,
    )

    # These are some default error handlers to use. Note the order matters here!
    # Only the first error handler triggered in this list will be used before
    # restarting the job
    error_handlers = [
        Unconverged(),
        NonConverging(),
        Frozen(),
        Walltime(),
    ]


# If you need to add advanced logic for one of your INCAR tags, you can register
# a keyword_modifier to the INCAR class like so:

# first define the logic of your modifier as a function
def keyword_modifier_multiply_nsites(structure, example_mod_input):
    # add your advanced logic to determine the keyword value.
    return structure.num_sites * example_mod_input


# then register modifier with the Incar class
Incar.add_keyword_modifier(keyword_modifier_multiply_nsites)

# -----------------------------------------------------------------------------
# That's it! Time to start building out your own workflows.
# -----------------------------------------------------------------------------



# -*- coding: utf-8 -*-

from pathlib import Path

from simmate.toolkit.diffusion import MigrationHop
from simmate.workflow_engine import Workflow
from simmate.calculators.vasp.workflows.diffusion.utilities import (
    get_migration_images_from_endpoints,
)
from simmate.calculators.vasp.workflows.relaxation.mvl_neb_endpoint import (
    Relaxation__Vasp__MvlNebEndpoint,
)
from simmate.calculators.vasp.workflows.diffusion.neb_from_images_mit import (
    Diffusion__Vasp__NebFromImagesMit,
)


class Diffusion__Vasp__NebSinglePathMit(Workflow):
    """
    Runs a full diffusion analysis on a single migration hop using NEB.

    Typically, this workflow is not ran directly -- but instead, you would
    use diffusion/all-paths which submits a series of this workflow for you.

    For a given migration path, supercell structures are generated for
    the start and end points of the migration, where vacancy-based diffsion
    is used. These supercells are relaxed using the relaxation/neb-endpoint
    workflow and then interpolated to generate 7 images. These 7 images are
    then relaxed using NEB within the diffusion/from-images workflow.

    This is therefore a "Nested Workflow" made of the following smaller workflows:

        - relaxation/neb-endpoint (for both start and end supercells)
        - a mini task that interpolated relaxed endpoints and makes images
        - diffusion/from-images

    Currently, this workflow will not run through the command-line, as we
    have not implemented a file format for MigrationHop's yet.
    """

    use_database = False
    description_doc_short = "runs NEB using a MigrationHop object as input"

    # TODO:
    # commands_out=["command_supercell", "command_neb"]

    @classmethod
    def run_config(
        cls,
        migration_hop: MigrationHop,
        directory: Path = None,
        source: dict = None,
        command: str = None,
        # These help link results to a higher-level table.
        diffusion_analysis_id: int = None,
        migration_hop_id: int = None,
        # TODO: Can the hop id be inferred from the migration_hop or somewhere
        # else in this context? Maybe even load_input_and_register will use
        # prefect id once it's a Calculation?
        is_restart: bool = False,
        # parameters for supercell and image generation
        nimages: int = 5,
        min_atoms: int = 80,
        max_atoms: int = 240,
        min_length: float = 10,
    ):

        # get the supercell endpoint structures
        supercell_start, supercell_end, _ = migration_hop.get_sc_structures(
            min_atoms=min_atoms,
            max_atoms=max_atoms,
            min_length=min_length,
            vac_mode=True,
        )

        # Relax the starting supercell structure
        endpoint_start_state = Relaxation__Vasp__MvlNebEndpoint.run(
            structure=supercell_start,
            command=command,  # subcommands["command_supercell"]
            directory=directory / f"{Relaxation__Vasp__MvlNebEndpoint.name_full}.start",
            is_restart=is_restart,
        )

        # Relax the ending supercell structure
        endpoint_end_state = Relaxation__Vasp__MvlNebEndpoint.run(
            structure=supercell_end,
            command=command,  # subcommands["command_supercell"]
            directory=directory / f"{Relaxation__Vasp__MvlNebEndpoint.name_full}.end",
            is_restart=is_restart,
        )

        # wait for the endpoint relaxations to finish
        endpoint_start_result = endpoint_start_state.result()
        endpoint_end_result = endpoint_end_state.result()

        images = get_migration_images_from_endpoints(
            supercell_start={
                "database_table": Relaxation__Vasp__MvlNebEndpoint.database_table.__name__,
                "directory": endpoint_start_result["directory"],
                "structure_field": "structure_final",
            },
            supercell_end={
                "database_table": Relaxation__Vasp__MvlNebEndpoint.database_table.__name__,
                "directory": endpoint_end_result["directory"],
                "structure_field": "structure_final",
            },
            nimages=nimages,
        )

        # Run NEB on this set of images
        neb_state = Diffusion__Vasp__NebFromImagesMit.run(
            migration_images=images,
            command=command,  # subcommands["command_neb"]
            source=source,
            directory=directory,
            diffusion_analysis_id=diffusion_analysis_id,
            migration_hop_id=migration_hop_id,
            is_restart=is_restart,
        )



# -*- coding: utf-8 -*-

import os
from pathlib import Path


def load_cluster_settings() -> dict:
    """
    Load our default cluster type and worker settings. This is loaded
    from the file `dask_cluster.yaml`.

    Note, module must be imported BEFORE we import any dask modules.
    """

    SIMMATE_DIRECTORY = Path.home() / "simmate"
    CLUSTER_YAML = SIMMATE_DIRECTORY / "dask_cluster.yaml"
    if CLUSTER_YAML.exists():
        os.environ.setdefault(
            "DASK_CONFIG",
            SIMMATE_DIRECTORY / "dask_cluster.yaml",
        )


# Automatically call the method above when importing this module
load_cluster_settings()



# -*- coding: utf-8 -*-

from simmate.conftest import copy_test_files, make_dummy_files
from simmate.calculators.vasp.inputs import Potcar
from simmate.calculators.vasp.workflows.electronic_structure.matproj_band_structure import (
    ElectronicStructure__Vasp__MatprojBandStructure,
)
from simmate.calculators.vasp.inputs.potcar_mappings import PBE_ELEMENT_MAPPINGS


def test_band_structure_setup(structure, tmp_path, mocker):

    # estabilish filenames that we make and commonly reference
    incar_filename = tmp_path / "INCAR"
    poscar_filename = tmp_path / "POSCAR"
    potcar_filename = tmp_path / "POTCAR"

    # Because we won't have POTCARs accessible, we need to cover this function
    # call -- specifically have it pretend to make a file
    mocker.patch.object(
        Potcar,
        "to_file_from_type",
        return_value=make_dummy_files(potcar_filename),
    )

    # try to make input files in the tmp_path
    ElectronicStructure__Vasp__MatprojBandStructure.setup(
        directory=tmp_path, structure=structure
    )
    assert incar_filename.exists()
    assert poscar_filename.exists()
    assert potcar_filename.exists()
    Potcar.to_file_from_type.assert_called_with(
        structure.composition.elements,
        "PBE",
        potcar_filename,
        PBE_ELEMENT_MAPPINGS,
    )


def test_band_structure_workup(tmp_path):
    copy_test_files(
        tmp_path,
        test_directory=__file__,
        test_folder="band_structure.zip",
    )

    # estabilish filenames that we make and commonly reference
    summary_filename = tmp_path / "simmate_summary.yaml"
    plot_filename = tmp_path / "band_structure.png"

    # run the full workup
    ElectronicStructure__Vasp__MatprojBandStructure.workup(tmp_path)
    assert summary_filename.exists()
    assert plot_filename.exists()



# -*- coding: utf-8 -*-

import time

# import pickle
import cloudpickle  # needed to serialize Prefect workflow runs and tasks

from simmate.workflow_engine.execution.database import WorkItem


class SimmateFuture:
    # class based on...
    # https://docs.python.org/3/library/concurrent.futures.html
    # Some methods still need to be added, but I have no need for them yet.
    # from concurrent.futures import Future # No need to inherit at the moment

    # locking rows is done by...
    # https://docs.djangoproject.com/en/3.1/ref/models/querysets/#select-for-update

    def __init__(self, pk):

        # This is the WorkItem personal key which tells us which row in the table
        # we should loook at.
        self.pk = pk

    def cancel(self):
        """
        Attempt to cancel the call. If the call is currently being executed or
        finished running and cannot be cancelled then the method will return
        False, otherwise the call will be cancelled and the method will return
        True.
        """
        # Query the WorkItem, lock it for editting, and check the status.
        workitem = WorkItem.objects.select_for_update().get(pk=self.pk)

        # check if the status is *not* PENDING
        if workitem.status != "P":
            # if so, the job is already running or finished, in which case
            # we can't cancel it.
            return False

        else:
            # If it is still pending, we can go ahead and cancel it.
            # This does not delete the task from the queue database though
            workitem.status = "C"
            workitem.save()
            return True

    def is_cancelled(self):
        """
        Return True if the call was successfully cancelled.
        """
        # I don't use a lock to check the status here
        workitem = WorkItem.objects.only("status").get(pk=self.pk)

        # check the status and indicate whether it is CANCELED or not
        if workitem.status == "C":
            return True
        else:
            return False

    def is_running(self):
        """
        Return True if the call is currently being executed and cannot be cancelled.
        """
        # I don't use a lock to check the status here
        workitem = WorkItem.objects.only("status").get(pk=self.pk)

        # check the status and indicate whether it is RUNNING or not
        if workitem.status == "R":
            return True
        else:
            return False

    def is_done(self):
        """
        Return True if the call was successfully cancelled or finished running.
        """
        # I don't use a lock to check the status here
        workitem = WorkItem.objects.only("status").get(pk=self.pk)

        # check the status and indicate whether it is FINISHED or CANCELED
        if workitem.status == "F" or workitem.status == "C":
            return True
        else:
            return False

    def result(self, timeout=None, sleep_step=0.1):
        """
        Return the value returned by the call. If the call hasn’t yet completed
        then this method will wait up to timeout seconds. If the call hasn’t
        completed in timeout seconds, then a concurrent.futures.TimeoutError
        will be raised. timeout can be an int or float. If timeout is not
        specified or None, there is no limit to the wait time.

        If the future is cancelled before completing then CancelledError
        will be raised.

        If the call raised, this method will raise the same exception.
        """
        # if no timeout was set, use infinity so we wait forever.
        if not timeout:
            timeout = float("inf")

        # Loop endlessly until the job completes or we timeout
        time_start = time.time()

        while (time.time() - time_start) < timeout:
            # I don't use a lock to check the status here
            workitem = WorkItem.objects.only("status", "result").get(pk=self.pk)
            status = workitem.status

            if status == "F":  # FINISHED
                # grab the result, unpickle it, and return it
                result = cloudpickle.loads(workitem.result)
                # if the result is an Error or Exception, raise it
                if isinstance(result, Exception):
                    raise result
                # otherwise return the result as-s
                else:
                    return result

            elif status == "C":  # CANCELED
                raise CancelledError("This item was cancelled and has no result")

            elif status == "P" or status == "R":  # PENDING or RUNNING
                # sleep the set amount before restarting the while loop
                time.sleep(sleep_step)

        # if the loop exits and we reached this line, then we've hit the timeout
        raise TimeoutError("The time-limit to wait for this result has been exceeded")


class CancelledError(Exception):
    pass



# -*- coding: utf-8 -*-

import logging
import importlib.metadata


__version__ = importlib.metadata.version("simmate")

# Some of these utilities use the __version__ above, so this import must come after
from simmate.utilities import get_doc_from_readme, check_if_using_latest_version

__doc__ = get_doc_from_readme(__file__)


# Configure our logger to output timestamps and "SIMMATE" with logs
# Also changes the logging level to info
logging.basicConfig(
    format="[SIMMATE-%(levelname)-s %(asctime)-s] %(message)s",
    level=logging.INFO,
    datefmt="%Y-%m-%d %H:%M:%S",
)


# Check if there is a newer Simmate version available, and if the current
# installation is out of date, print a message for the user.
try:
    check_if_using_latest_version()
except:
    logging.warn("Unable to check if using the latest Simmate version.")



# -*- coding: utf-8 -*-

from simmate.calculators.vasp.workflows.relaxation.matproj_scan import (
    Relaxation__Vasp__MatprojScan,
)


class StaticEnergy__Vasp__MatprojScan(Relaxation__Vasp__MatprojScan):
    """
    This task is a reimplementation of pymatgen's
    [MPScanStaticSet](https://pymatgen.org/pymatgen.io.vasp.sets.html#pymatgen.io.vasp.sets.MPScanStaticSet).
    """

    incar = Relaxation__Vasp__MatprojScan.incar.copy()
    incar.update(
        dict(
            NSW=0,  # this is the main static energy setting
            LREAL=False,
            LORBIT=11,
            LVHAR=True,
            ISMEAR=-5,
        )
    )



# -*- coding: utf-8 -*-

"""
# Example use

``` python
from simmate.workflows.customized import Customized__Vasp__UserConfig as workflow

result = workflow.run(
    workflow_base="relaxation.vasp.quality04",
    input_parameters={"structure": "NaCl.cif"},
    updated_settings={
        "incar": {"NPAR": 1, "ENCUT": 600},
    },
).result()
```

# Example input YAML

``` yaml
# Indicates we want to change the settings, using a specific workflow as a starting-point
workflow_name: vasp/customized
base_workflow: static-energy/mit

# These would update the class attributes for the single workflow run
# The "custom_" start indicates we are updating some attribute
custom_incar: 
    - ENCUT: 600
    - KPOINTS: 0.25
custom_potcar_mappings:
    - Y: Y_sv

# Then the remaining inputs are the same as the base_workflow
structure: POSCAR
command: mpirun -n 5 vasp_std > vasp.out
```
"""

import logging

from simmate.workflow_engine import Workflow, S3Workflow


class Customized__Vasp__UserConfig(Workflow):
    """
    "VASP calculation with updated INCAR/POTCAR settings
    """

    @staticmethod
    def run_config(
        workflow_base: Workflow,
        updated_settings: dict,
        input_parameters: dict,
    ):
        logging.warn(
            "WARNING: customized workflows are meant only for quick testing. "
            "If you are using custom settings regularly, we highly recommend "
            "making a new workflow inside a Simmate project instead. "
            "Read through tutorial 06 for more information."
        )

        # ensure this workflow has a S3Task class attribute
        if not issubclass(workflow_base, S3Workflow):
            raise Exception(
                "Dynamically updating settings is only supported for S3Workflows"
            )

        # we can't just copy the workflow and update attributes, as this will update
        # attributes of source workflow as well. Instead, we must subclass the
        # workflow and then update its attributes. Therefore, this process below seeks
        # to copy and update relevant attributes before subclassing at the end.
        #   see https://stackoverflow.com/questions/9541025/

        # we start with the original s3workflow class and keep track of
        # attributes we need to update.
        final_attributes = {}

        # Key will be something like "incar" and the update_values will be a
        # dictionary such as {"ENCUT": 520, "NPAR": 3}
        for update_attribute, update_values in updated_settings.items():

            original_config = getattr(workflow_base, update_attribute)

            # we will be deleting/editting the dictionary so we need to make a copy
            new_config = original_config.copy()

            # If a user sets EDIFF = 1e-3 and there is EDIFF__per_atom in the
            # incar dictionary, we need to remove that key as well. We therefore
            # remove all other "examplekey__*" settings from the dictionary.
            for update_key in update_values.keys():
                for subkey in original_config.keys():
                    if subkey.startswith(f"{update_key}__"):
                        new_config.pop(subkey)
            # BUG: what about keywords like multiple_keywords__smart_ldau? I may need
            # to refactor the INCAR class to handle this.

            # now we can add the custom values
            new_config.update(update_values)

            final_attributes[update_attribute] = new_config

        # As an extra, we disable any registration & saving of results to the database
        final_attributes["use_database"] = False

        # Dynamically create a subclass of the original s3task
        NewWorkflow = type(
            f"{workflow_base.__name__}Custom", (workflow_base,), final_attributes
        )

        # now run the task with the remaining parameters
        # Note, we are using run_config instead of run because we do not want this
        # registered as a task run of the original class
        state = NewWorkflow.run(**input_parameters)
        result = state.result()

        return result



# -*- coding: utf-8 -*-

from simmate.calculators.vasp.workflows.relaxation.matproj import (
    Relaxation__Vasp__Matproj,
)


class Elastic__Vasp__Mvl(Relaxation__Vasp__Matproj):
    """
    This task is a reimplementation of pymatgen's
    [MVLElasticSet](https://pymatgen.org/pymatgen.io.vasp.sets.html#pymatgen.io.vasp.sets.MVLElasticSet).
    """

    incar = Relaxation__Vasp__Matproj.incar.copy()
    incar.update(
        dict(
            IBRION=6,
            NFREE=2,
            POTIM=0.015,  # pymatgen uses this as a parameter
        )
    )
    # incar.pop("NPAR")  # pymatgen forcibly removes NPAR



# -*- coding: utf-8 -*-

from simmate.conftest import copy_test_files
from simmate.calculators.vasp.workflows.electronic_structure.matproj_density_of_states import (
    ElectronicStructure__Vasp__MatprojDensityOfStates,
)


def test_density_of_states_workup(tmp_path):
    copy_test_files(
        tmp_path,
        test_directory=__file__,
        test_folder="density_of_states.zip",
    )

    # estabilish filenames that we make and commonly reference
    summary_filename = tmp_path / "simmate_summary.yaml"
    plot_filename = tmp_path / "density_of_states.png"

    # run the full workup
    ElectronicStructure__Vasp__MatprojDensityOfStates.workup(tmp_path)
    assert summary_filename.exists()
    assert plot_filename.exists()



# -*- coding: utf-8 -*-

"""
This module defines the base class for all workflows in Simmate. When learning 
how use workflows, make sure you have gone through our intro 
[tutorials](https://github.com/jacksund/simmate/tree/main/tutorials). You
can then read through these guides for more features.



# Using existing workflows


## Loading a workflow

To import a specific workflow, see the `simmate.workflows` module. For all the
examples below, we will look at the `static-energy.vasp.matproj` workflow,
which we load with:
    
``` python
from simmate.workflows.all import StaticEnergy__Vasp__Matproj as workflow
```

Alternatively, you can also use `get_workflow` to load your workflow:
    
``` python
from simmate.workflows.utilities import get_workflow

workflow = get_workflow("static-energy.vasp.matproj")
```


## Viewing input parameters

A workflow's nput parameters can be analyzed in the following ways:

``` python
workflow.show_parameters()  # prints out info
workflow.parameter_names
workflow.parameter_names_required
```


## Running a workflow (local)

To run a workflow locally (i.e. directly on your current computer), you can
use the `run` method. This is returns a `State` object, which allows
you to check if the run completed successfully or not. Then final output of your
workflow run can be accessed using `state.result()`. You can read more about
`State` objects [here](https://orion-docs.prefect.io/concepts/states/), but as
a quick example:

``` python
state = workflow.run(
    structure="NaCl.cif", 
    command="mpirun -n 4 vasp_std > vasp.out",
)

if state.is_completed():  # optional check
    result = state.result()
```

Outside of python, you can also run workflows from the command line, using YAML 
files, or the website interface. These approaches are covered in the intro
tutorials.


## Running a workflow (cloud)

When you want to schedule a workflow to run elsewhere, you must first make sure
you have your computational resources configured. You can then run workflows
using the `run_cloud` method, which returns a Prefect flow run id.

``` python
run_id = workflow.run_cloud(
    structure="NaCl.cif", 
    command="mpirun -n 4 vasp_std > vasp.out",
)
```

On the computer where you want to actually run the workflow, start a worker
that will pick up and run the workflow you just scheduled:

``` bash
# NOTE: This worker will run endlessly.
# To control shutdown, see the Worker documentation.
simmate workflow-engine start-worker
```

If you want better control of which workers select a given workflow, you can
use `tags` to label your submission and worker.

``` python
run_id = workflow.run_cloud(
    structure="NaCl.cif", 
    tags=["my-custom-tag1", "quick-job"]
)
```

``` bash
simmate workflow-engine start-worker -t my-custom-tag1 -t quick-job
```


## Accessing results

In addition to using `state.result()` as described above, you can also view
results database through the `database_table` attribute (if one is available).
This returns a Simmate database object for results of ALL runs of this workflow.
Guides for filtering and manulipating the data in this table is covered in
the `simmate.database` module guides. But as an example:

``` python
table = workflow.database_table

# pandas dataframe that you can view in Spyder
df = table.obects.to_dataframe()

# or grab a specific run result and convert to a toolkit object
entry = table.objects.get(run_id="example-123456")
structure = entry.to_toolkit()
```

Filtering of results is covered in the `simmate.database` documentation.

<!--
TODO:
    - viewing parameter names
    - accessing results (database_table)
    - run_cloud and other cloud methods
    - different documentation levels (parameters, mini docstring, etc.)
    - finding the same workflow in the website UI (place this )
-->




# Creating new workflows

## Naming workflows (required)

Higher level features such as the website interface require that workflow
names follow a certain format. If you skip this step, your workflows
will fail and cause errors elsewhere.

First, we need to update the workflow name to match Simmate's naming
conventions, which includes:
    1.  The type of analysis the workflow is doing
    2.  The "calculator" (or program) that the workflow uses to run
    3.  A unique name to identify the settings used

Examples for each part would be:
    1. relaxation, static-energy, dynamics, ...
    2. vasp, abinit, qe, deepmd, ...
    3. jacks-test, matproj, quality00, ...

Together, an example workflow names would be:
    - `relaxation.vasp.jacks-test`
    - `static-energy.abinit.matproj`
    - `dynamics.qe.quality00`

When converting this to our workflow name in python, we need to replace
periods with 2 underscores each and convert our words to
[pascal case](https://khalilstemmler.com/blogs/camel-case-snake-case-pascal-case/).
For example, our workflow names become:
    - `Relaxation__Vasp__JacksTest`
    - `StaticEnergy__Abinit__Matproj`
    - `Dynamics__Qe__Quality00`

NOTE: Capitalization is very important here so make sure you double check
your workflow names.

Now let's test this out in python using a similar workflow name:
``` python
from simmate.workflow_engine import Workflow

class Example__Python__MyFavoriteSettings(Workflow):
    pass  # we will build the rest of workflow later

# These names can be long and unfriendly, so it can be nice to
# link them to a variable name for easier access.
my_workflow = Example__Python__MyFavoriteSettings

# Now check that our naming convention works as expected
assert my_workflow.name_full == "example.pure-python.my-favorite-settings"
assert my_workflow.name_type == "example"
assert my_workflow.name_calculator == "python"
assert my_workflow.name_preset == "my-favorite-settings"
```

You now have a ready-to-use workflow name!


## Example workflow

To convert this to a Simmate workflow, we just need to change the format
a little. Instead of a `@flow` decorator, we use the `run_config` method
of a new subclass:

``` python
# NOTE: this example does not follow Simmate's naming convention, so
# some higher level features will be broken. We will fix in a later step.

from simmate.workflow_engine import Workflow

class Example__Python__MyFavoriteSettings(Workflow):
    
    use_database = False  # we don't have a database table yet

    @staticmethod
    def run_config():
        print("This workflow doesn't do much")
        return 42

# and then run your workflow
state = MyFavoriteWorkflow.run()
result = state.result()
```

Behind the scenes, the `run` method is converting our `run_config` to a 
workflow for us.

Now let's look at a realistic example where we build a Workflow that has
input parameters and accesses class attributes/methods:

``` python

class Example__Python__MyFavoriteSettings(Workflow):
    
    use_database = False # we don't have a database table yet
    example_constant = 12

    @staticmethod
    def squared(x):
        return x ** 2

    @classmethod
    def run_config(cls, name, say_hello=True, **kwargs):
        # Workflows can contain ANY python code!
        # In other words...
        #   "The ceiling is the roof" -Michael Jordan

        if say_hello:
            print(f"Hello and welcome, {name}!")

        # grab class values and methods
        x = cls.example_constant
        example_calc = cls.squared(x)
        print(f"Our calculation gave a result of {example_calc}")

        # grab extra arguments if you need them
        for key, value in kwargs.items():
            print(
                f"An extra parameter for {key} was given "
                "with a value of {value}"
            )

        return "Success!"
```

Once you have your new workflow, you can run it as you would any other worfklow:

``` python
state = Example__Python__MyFavoriteSettings.run("Jack")
result = state.result()
```


## Modifying an existing workflow

You can use python inheritance to borrow utilities and settings from workflows.
However, make sure you are making copies of the original workflow settings!
If you modify them without making a copy, you'll be in a world of trouble.

``` python
from simmate.workflows.utilities import get_workflow

original_workflow = get_workflow("static-energy.vasp.matproj")


class StaticEnergy__Vasp__MyCustomPreset(original_workflow):

    version = "2022.07.04"

    incar = original_workflow.incar.copy()  # Make sure you copy!
    incar.update(
        dict(
            NPAR=1,
            ENCUT=-1,
        )
    )

# make sure we have new settings updated
# and that we didn't change the original
assert original_workflow.incar != StaticEnergy__Vasp__MyCustomPreset
```


## Linking a database table

Many of workflows will want to store common types of data (such as StaticEnergy
or Relaxation data). If you would like to use these tables automatically, you 
simply to make sure you `name_type` matches what is available!

For example, if we look at a static-energy calculation, you will see
the `StaticEnergy` database table is automatically used because the
name of our workflow starts with "StaticEnergy":

``` python
from simmate.database import connect
from simmate.database.workflow_results import StaticEnergy

# no work required! This line shows everything is setup and working
assert StaticEnergy__Vasp__MyCustomPreset.database_table == StaticEnergy
```

If you would like to build or use a custom database, you must first have
a registered `DatabaseTable` (this is covered in the main tutorials), and
then you can link the database table to your workflow directly. The only
other requiredment is that your database table uses the `Calculation`
database mix-in: 

``` python

from my_project.models import MyCustomTable

class Example__Python__MyFavoriteSettings(Workflow):
    database_table = MyCustomTable
```


## Registering your workflow

Registering your workflow so that you can access it in the UI requires you to
build a "simmate project". This is covered in the main tutorials.


## Common workflow types

In many cases, you may have a workflow that runs a single calculation
and then writes the results to a specific database table. An example of
this would be an energy calculation using VASP. If your workflow involves
calling another program, you should read about the `S3Workflow` which helps
with writing input files, calling other programs, and handling errors.
See the documentation for `simmate.workflow_engine.s3_workflow`.


# Prefect backend (experimental)

When you enable Prefect as your workflow executor, Workflows are converted into
Prefect Flows under the hood, so having knowledge of Prefect can be useful in 
advanced cases. For advanced use or when building new features, we recommend 
going through the prefect tutorials located
[here](https://orion-docs.prefect.io/tutorials/first-steps/).


## Minimal example (Prefect vs. Simmate)

It's useful to know how Prefect workflows compare to Simmate workflows. For
simple cases where you have python code, you'd define a Prefect workflow
like so:

``` python
from prefect import flow

@flow
def my_favorite_workflow():
    print("This workflow doesn't do much")
    return 42

# and then run your workflow
state = my_favorite_workflow()
result = state.result()
```

To convert this to a Simmate workflow, we just need to change the format
a little. Instead of a `@flow` decorator, we use the `run_config` method
of a new subclass:

``` python
# NOTE: this example does not follow Simmate's naming convention, so
# some higher level features will be broken. We will fix in a later step.

from simmate.workflow_engine import Workflow

class Example__Python__MyFavoriteSettings(Workflow):

    @staticmethod
    def run_config():
        print("This workflow doesn't do much")
        return 42

# and then run your workflow
state = MyFavoriteWorkflow.run()
result = state.result()
```

Behind the scenes, the `run` method is converting our `run_config` to a
Prefect workflow for us. Methods like `run_cloud` will automatically use
Prefect now too.

"""

import logging
import json
import cloudpickle
import yaml
import re
import inspect
from typing import List, Any
import uuid
from pathlib import Path

import simmate
from simmate.database.base_data_types import Calculation
from simmate.utilities import get_directory, copy_directory
from simmate.workflow_engine.execution import SimmateExecutor, WorkItem


class DummyState:
    """
    This class is meant to emulate Prefect States. By wrapping a result into
    State, we enable higher-level features that depend on a call to
    `state.result()`.

    This class should not be used directly as it is automatically applied with
    the `Workflow.run` method
    """

    def __init__(self, result):
        self._result = result

    def result(self):
        return self._result

    @staticmethod
    def is_completed():
        return True


class Workflow:
    """
    An abstract base class for all Simmate workflows.
    """

    # TODO: set storage attribute to module

    version: str = simmate.__version__
    """
    Version number for this flow. Defaults to the Simmate version 
    (e.g. "0.7.0").
    """

    description_doc_short: str = None
    """
    A quick description for this workflow. This will be shown in the website UI
    in the list-view of all different workflow presets.
    """

    use_database: bool = True
    """
    Whether to use Simmate database features or not.
    
    This includes calling the `_register_calculation` and `_save_to_database`
    methods attached to this workflow.
    
    `_register_calculation` will save a database entry before the workflow
    starts. This is useful to keep track of workflows that have been
    submitted/started but haven't finished yet.
    
    `_save_to_database` saves the output of the `workup` method to the database.
    """

    _parameter_methods = ["run_config"]
    """
    List of methods that allow unique input parameters. This helps track where
    `**kwargs` are passed and let's us gather the inputs in one place.
    """

    # -------------------------------------------------------------------------
    # Core methods that handle how and what a workflow run does
    # and how it is submitted
    # -------------------------------------------------------------------------

    @classmethod
    def run_config(cls):
        """
        The workflow method, which can be overwritten when inheriting from this
        class. This can be either a staticmethod or classmethod.
        """
        raise NotImplementedError(
            "When creating a custom workflow, make sure you set a run_config method!"
        )

    @classmethod
    def _run_full(cls, run_id=None, **kwargs):
        """
        This method should not be called directly. Use the `run` method instead.
        """
        logging.info(f"Starting {cls.name_full}")
        # This method is isolated only because we want to wrap it as a prefect
        # workflow in some cases.
        run_id = run_id or cls._get_run_id()
        kwargs_cleaned = cls._load_input_and_register(run_id=run_id, **kwargs)
        result = cls.run_config(**kwargs_cleaned)
        if cls.use_database:
            result["calculation_id"] = cls._save_to_database(result, run_id=run_id)
        logging.info(f"Completed {cls.name_full}")
        return result

    @staticmethod
    def _get_run_id():
        """
        generates a random id to use as a workflow run id
        """
        # This is a separate method in order to allow the prefect executor to
        # overwrite this method.
        unique_id = str(uuid.uuid4())
        return unique_id

    @classmethod
    def run(cls, **kwargs) -> DummyState:
        """
        runs the workflow locally
        """
        # Note: this is a separate method and wrapper around run_full because
        # we want to allow Prefect executor to overwrite this method.
        result = cls._run_full(**kwargs)
        state = DummyState(result)
        return state

    @classmethod
    def run_cloud(
        cls, return_state: bool = True, tags: list[str] = None, **kwargs
    ) -> str:
        """
        submits the workflow run to cloud database to be ran by a worker
        """

        logging.info(f"Submitting new run of {cls.name_full} to cloud")

        # If we are submitting using a filename, we don't want to
        # submit to a cluster and have the job fail because it doesn't have
        # access to the file. We therefore deserialize right before
        # serializing in the next line in order to ensure parameters that
        # accept file names are submitted with all necessary data.
        parameters_deserialized = cls._deserialize_parameters(**kwargs)
        parameters_serialized = cls._serialize_parameters(**parameters_deserialized)

        # Because we often want to save some info to our database even before
        # the calculation starts/finishes, we do that by calling _register_calc
        # at this higher level. An example is storing the structure and run id.
        # Thus, we create and register the run_id up front
        run_id = cls._get_run_id()
        if cls.use_database:
            cls._register_calculation(run_id=run_id, **kwargs)

        state = SimmateExecutor.submit(
            cls._run_full,  # should this be the run method...?
            run_id=run_id,
            tags=tags or cls.tags,
            **parameters_serialized,
        )

        logging.info(f"Successfully submitted (run_id={run_id})")

        # If the user wants the future, return that instead of the run_id
        if return_state:
            state.run_id = run_id  # attach the run id as an extra
            return state

        return run_id

    # -------------------------------------------------------------------------
    # Methods that interact with the Executor class in order to see what
    # has been submitted to cloud.
    # -------------------------------------------------------------------------

    @classmethod
    @property
    def nflows_submitted(cls) -> int:
        """
        Queries the Simmate database to see how many workflows are in a
        running or pending state.
        """
        return WorkItem.objects.filter(status__in=["P", "R"], tags=cls.tags).count()

    # -------------------------------------------------------------------------
    # Methods that help with accessing the database and saving results
    # -------------------------------------------------------------------------

    @classmethod
    @property
    def database_table(cls) -> Calculation:
        """
        The database table where calculation information (such as the run_id)
        is stored. The table should use `simmate.database.base_data_types.Calculation`

        In many cases, this table will contain all of the results you need. However,
        pay special attention to NestedWorkflows, where your results are often tied
        to a final task.
        """
        flow_type = cls.name_type
        flow_preset = cls.name_preset

        if flow_type == "relaxation":
            from simmate.database.base_data_types import Relaxation

            return Relaxation
        elif flow_type == "static-energy":
            from simmate.database.base_data_types import StaticEnergy

            return StaticEnergy
        elif flow_type == "electronic-structure":
            if "band-structure" in flow_preset:
                from simmate.database.base_data_types import BandStructureCalc

                return BandStructureCalc
            elif "density-of-states" in flow_preset:
                from simmate.database.base_data_types import (
                    DensityofStatesCalc,
                )

                return DensityofStatesCalc
        elif flow_type == "population-analysis":
            from simmate.database.base_data_types import PopulationAnalysis

            return PopulationAnalysis
        elif flow_type == "dynamics":
            from simmate.database.base_data_types import DynamicsRun

            return DynamicsRun
        elif flow_type == "diffusion":
            if "from-images" in flow_preset:
                from simmate.database.base_data_types import MigrationImage

                return MigrationImage
            else:
                from simmate.database.base_data_types import DiffusionAnalysis

                return DiffusionAnalysis
        elif flow_type == "customized":
            from simmate.database.base_data_types import CustomizedCalculation

            return CustomizedCalculation
        else:
            raise NotImplementedError("Unable to detect proper database table")

    @classmethod
    def _save_to_database(cls, result, run_id):

        # split our results and corrections (which are given as a dict) into
        # separate variables
        vasprun = result["result"]
        corrections = result["corrections"]
        directory = result["directory"]

        # load the calculation entry for this workflow run. This should already
        # exist thanks to the load_input_and_register task.
        calculation = cls.database_table.from_run_context(
            run_id=run_id,
            workflow_name=cls.name_full,
        )

        # now update the calculation entry with our results
        calculation.update_from_vasp_run(vasprun, corrections, directory)

        return calculation.id

    # -------------------------------------------------------------------------
    # Properties that enforce the naming convention for workflows
    # -------------------------------------------------------------------------

    @classmethod
    @property
    def name_full(cls) -> str:
        """
        Standardized name of the workflow. This converts the class name like so:
        `Static_Energy__VASP__Matproj` --> `static-energy.vasp.matproj`
        """
        if not len(cls.__name__.split("__")) == 3:
            raise Exception("Make sure you are following Simmate naming conventions!")

        # convert to dot format
        name = cls.__name__.replace("__", ".")

        # adds a hyphen between each capital letter
        # copied from https://stackoverflow.com/questions/199059/
        name = re.sub(r"(\w)([A-Z])", r"\1-\2", name)

        return name.lower()

    @classmethod
    @property
    def name_type(cls) -> str:
        """
        Name of the Project this workflow is associated with. This is the first
        portion of the flow name (e.g. "static-energy")
        """
        return cls.name_full.split(".")[0]

    @classmethod
    @property
    def name_calculator(cls) -> str:
        """
        Name of the calculator this workflow is associated with. This is the second
        portion of the flow name (e.g. "vasp")
        """
        return cls.name_full.split(".")[1]

    @classmethod
    @property
    def name_preset(cls) -> str:
        """
        Name of the settings/preset this workflow is associated with. This is the third
        portion of the flow name (e.g. "matproj" or "matproj-prebader")
        """
        return cls.name_full.split(".")[2]

    @classmethod
    @property
    def tags(cls) -> list[str]:
        """
        Lists of tags to submit a the workflow with when using run_cloud.
        """
        return [
            "simmate",
            cls.name_type,
            cls.name_calculator,
            cls.name_full,
        ]

    # -------------------------------------------------------------------------
    # Properties/method that set website UI documentation and help users
    # explore input options and settings used.
    # -------------------------------------------------------------------------

    # BUG: naming this `description` causes issues.
    # See https://github.com/PrefectHQ/prefect/issues/3911
    @classmethod
    @property
    def description_doc(cls) -> str:
        """
        This simply returns the documentation string of this workflow -- so this
        is the same as `__doc__`. This attribute is only defined for beginners
        to python and for use in django templates for the website interface.
        """
        return cls.__doc__

    @classmethod
    @property
    def parameter_names(cls) -> List[str]:
        """
        Gives a list of all the parameter names for this workflow.
        """
        # Iterate through and grab the parameters for the core methods. We also
        # sort parameters alphabetically for consistent results.
        parameters = []
        for method in cls._parameter_methods:
            sig = inspect.signature(getattr(cls, method))
            for parameter in list(sig.parameters):
                if parameter not in parameters and parameter != "kwargs":
                    parameters += [parameter]
        parameters.sort()
        return parameters

    @classmethod
    @property
    def parameter_names_required(cls) -> List[str]:
        """
        Gives a list of all the required parameter names for this workflow.
        """
        parameters = []
        for method in cls._parameter_methods:
            sig = inspect.signature(getattr(cls, method))
            for parameter, val in sig.parameters.items():
                if (
                    parameter not in parameters
                    and parameter not in ["directory", "kwargs"]
                    and val.default == val.empty
                ):
                    parameters += [parameter]
        parameters.sort()
        return parameters

    @classmethod
    def show_parameters(cls):
        """
        Prints a list of all the parameter names for this workflow.
        """
        # use yaml to make the printout pretty (no quotes and separate lines)
        print(yaml.dump(cls.parameter_names))

    @classmethod
    def get_config(cls):
        """
        Grabs the overall settings from the class in order to let users briefly
        look at key parameters.

        Using this method is scripts is not recommended.
        """
        return dict(
            use_database=cls.use_database,
            name_full=cls.name_full,
            parameter_names=cls.parameter_names,
            parameter_names_required=cls.parameter_names_required,
        )

    @classmethod
    def show_config(cls):
        """
        Takes the result of get_config and prints it in a yaml format that is
        easier to read.
        """
        config = cls.get_config()
        print(yaml.dump(config))

    # -------------------------------------------------------------------------
    # Properties/method that configure registration of workflow runs and the
    # parameters each is called with.
    # -------------------------------------------------------------------------

    @classmethod
    def _load_input_and_register(cls, run_id: str, **parameters: Any) -> dict:
        """
        How the input was submitted as a parameter depends on if we are submitting
        to Prefect Cloud, running the flow locally, or even continuing from a
        previous calculation.  Here, we use a task to convert the input to a toolkit
        object and (if requested) provide the directory as well.

        For example, a common input parameter for workflows is "structure", which
        can be provided a number of ways:
            - a filename
            - a json string
            - a dictionary pointing to a database entry
            - a toolkit Structure object
            - etc...
        Even though all of these inputs are accepted, `workflow.run` always expects
        python objects, so this utility converts the input to a toolkit Structure
        object.

        `register_run` allows us to skip the database step if the database_table
        isn't properly set yet. This input is a temporary fix for the
        diffusion/from-images workflow.

        `copy_previous_directory` is only used when we are pulling a structure from a
        previous calculation. If copy_previous_directory=True, then the directory
        parameter is ignored.

        `**parameters` includes all parameters and anything extra that you want saved
        to simmate_metadata.yaml
        """

        # OPTIMIZE: consider splitting this task into load_structure, load_directory,
        # and register_calc so that our flow_visualize looks cleaner

        # OPTIMIZE: Customized workflows cause a lot of special handling in this task
        # so it may be worth isolating these into a separate task.

        # ---------------------------------------------------------------------

        # STEP 1: clean parameters

        parameters_cleaned = cls._deserialize_parameters(**parameters)

        # ---------------------------------------------------------------------

        # STEP 1b: Determine the "primary" input to use for setting the
        # source (and previous directory)
        # OPTIMIZE: Is there a better way to do this?

        # Currently I just set a priority of possible parameters that can be
        # the primary input. I go through each one at a time until I find one
        # that was provided -- then I exit with that parameter's value.
        primary_input = None
        for primary_input_key in [
            "structure",
            "migration_hop",
            "supercell_start",
        ]:
            primary_input = parameters.get(primary_input_key, None)
            primary_input_cleaned = parameters_cleaned.get(primary_input_key, None)
            if primary_input:
                break

        # ---------------------------------------------------------------------

        # STEP 2: Load the directory (and copy over from an old directory if necessary)

        # Start by creating a new directory or grabbing the one given. We create
        # this directory immediately (rather than just passing the name to the
        # S3Task). We do this because NestedWorkflows often use a parent directory
        # to organize results.
        directory = parameters.get("directory", None)
        directory_cleaned = get_directory(directory)

        # if the user requested, we grab the previous directory as well
        copy_previous_directory = parameters.get("copy_previous_directory", None)
        if copy_previous_directory:

            if not primary_input:
                raise Exception(
                    "No primary input detected, which is required for copying "
                    "past directories. This is an experimental feature so "
                    "please contact our team for more help."
                )

            # catch incorrect use of this function
            if not primary_input_cleaned.is_from_past_calc:
                raise Exception(
                    "There isn't a previous directory available! Your source "
                    "structure must point to a past calculation to use this feature."
                )

            # the past directory should be stored on the input object
            previous_directory = Path(primary_input_cleaned.database_object.directory)

            # Copy over all files except simmate ones (we have no need for the
            # summaries or error archives)
            copy_directory(
                directory_old=previous_directory,
                directory_new=directory_cleaned,
                ignore_simmate_files=True,
            )

        # SPECIAL CASE for customized flows
        if "workflow_base" not in parameters_cleaned:
            parameters_cleaned["directory"] = directory_cleaned
        else:
            parameters_cleaned["input_parameters"]["directory"] = directory_cleaned

        # ---------------------------------------------------------------------

        # STEP 3: Load the source of the input object

        source = parameters.get("source", None)

        # If we were given a input from a previous calculation, the source should
        # point directory to that same input. Otherwise we are incorrectly trying
        # to change what the source is.
        # "primary_input and" is added to the start to ensure cleaned input exists
        # and therefore prevent an error/bug.
        if source and primary_input and primary_input_cleaned.is_from_past_calc:
            # note primary_input here is a dictionary
            # assert
            if not source == primary_input:
                # only warning for now because this is experimental
                logging.warn(
                    "Your source does not match the source of your "
                    "primary input. Sources are an experimental feature, so "
                    "this will not affect your results. Still, please report "
                    "this to our team to help with development. \n\n"
                    f"SOURCE: {source} \n\n"
                    f"PRIMARY_INPUT: {primary_input} \n\n"
                )
            source_cleaned = source
        # Check if we have a primary input loaded from a past calculation and
        # default to that as the source.
        elif primary_input and primary_input_cleaned.is_from_past_calc:
            source_cleaned = primary_input
        # Otherwise just use the source given
        elif source:
            source_cleaned = source
        else:
            source_cleaned = None

        # SPECIAL CASE for customized flows
        if "workflow_base" not in parameters_cleaned:
            parameters_cleaned["source"] = source_cleaned
        else:
            parameters_cleaned["input_parameters"]["source"] = source_cleaned

        # ---------------------------------------------------------------------

        # STEP 4: Register the calculation so the user can follow along in the UI
        # and also see which structures/runs have been submitted aready.

        if cls.use_database:
            cls._register_calculation(run_id=run_id, **parameters_cleaned)

        # ---------------------------------------------------------------------

        # STEP 5: Write metadata file for user reference

        # convert back to json format. We convert back rather than use the original
        # to ensure the input data is all present. For example, we want to store
        # structure data instead of a filename in the metadata.
        # SPECIAL CASE: "if ..." used to catch customized workflows
        parameters_serialized = (
            cls._serialize_parameters(**parameters_cleaned)
            if "workflow_base" not in parameters_cleaned
            else parameters
        )

        # We want to write a file summarizing the inputs used for this
        # workflow run. This allows future users to reproduce the results if
        # desired -- and it also allows us to load old results into a database.
        input_summary = dict(
            workflow_name=cls.name_full,
            # this ID is ingored as an input but needed for loading past data
            run_id=run_id,
            **parameters_serialized,
        )

        # now write the summary to file in the same directory as the calc.
        input_summary_filename = directory_cleaned / "simmate_metadata.yaml"
        with input_summary_filename.open("w") as file:
            content = yaml.dump(input_summary)
            file.write(content)

        # ---------------------------------------------------------------------

        # Finally we just want to return the dictionary of cleaned parameters
        # to be used by the workflow
        return parameters_cleaned

    @classmethod
    @property
    def _parameters_to_register(cls) -> List[str]:
        """
        A list of input parameters that should be used to register the calculation.
        """

        # run is always used to register but is never an input parameter
        parameters_to_register = []
        # run_id and workflow_name are used to register but these are
        # implemented with the _register_calculation elsewhere

        table_columns = cls.database_table.get_column_names()

        for parameter in cls.parameter_names:
            if parameter in table_columns:
                parameters_to_register.append(parameter)

        # check special cases where input parameter doesn't match to a column name
        if "structure_string" in table_columns:
            parameters_to_register.append("structure")

        # put in alphabetical order for consistent results
        parameters_to_register.sort()

        return parameters_to_register

    @classmethod
    def _register_calculation(cls, run_id=None, **kwargs) -> Calculation:
        """
        If the workflow is linked to a calculation table in the Simmate database,
        this adds the flow run to the database.

        Parameters passed should be deserialized and cleaned.

        This method should not be called directly as it is used within the
        `run_prefect_cloud` method and `load_input_and_register` task.
        """

        # We first need to grab the database table where we want to register
        # the calculation run to. We can grab the table from either...
        #   1. the database_table attribute
        #   2. flow_context --> flow_name --> flow --> then grab its database_table

        # If this method is being called on the base Workflow class, that
        # means we are trying to register a calculation from within a flow
        # context -- where the context has information such as the workflow
        # we are using (and the database table linked to that workflow).
        if cls == Workflow:
            raise Exception("Checking if this method is ever used")

            from prefect.context import FlowRunContext

            run_context = FlowRunContext.get()
            workflow = run_context.flow.simmate_workflow
            database_table = workflow.database_table

        # Otherwise we should be using the subclass Workflow that has the
        # database_table property set.
        else:
            workflow = cls  # we have the workflow class already
            database_table = cls.database_table

        # Registration is only possible if a table is provided. Some
        # special-case workflows don't store calculation information bc the flow
        # is just a quick python analysis.
        if not database_table:
            logging.warn("No database table found. Skipping registration.")
            return

        # grab the registration kwargs from the parameters provided and then
        # convert them to a python object format for the database method
        register_kwargs = {
            key: kwargs.get(key, None) for key in workflow._parameters_to_register
        }
        register_kwargs_cleaned = cls._deserialize_parameters(
            add_defaults_from_attr=False, **register_kwargs
        )

        # SPECIAL CASE: for customized workflows we need to convert the inputs
        # back to json before saving to the database.
        if "workflow_base" in register_kwargs_cleaned:
            parameters_serialized = cls._serialize_parameters(**register_kwargs_cleaned)
            calculation = database_table.from_run_context(
                run_id=run_id,
                workflow_name=cls.name_full,
                **parameters_serialized,
            )
        else:
            # load/create the calculation for this workflow run
            calculation = database_table.from_run_context(
                run_id=run_id,
                workflow_name=cls.name_full,
                **register_kwargs_cleaned,
            )

        return calculation

    @classmethod
    def _serialize_parameters(cls, **parameters) -> dict:
        """
        Converts input parameters to json-sealiziable objects that Prefect can
        use.

        This method should not be called directly as it is used within the
        run_prefect_cloud() method.
        """

        # TODO: consider moving this into prefect's core code as a contribution.
        # This alternatively might be a pydantic contribution

        # Because many flows allow object-type inputs (such as structure object),
        # we need to serialize these inputs before scheduling them with prefect
        # cloud. To do this, I only check for two potential methods:
        #     as_dict
        #     to_dict
        # These were chosen based on common pymatgen methods, but I may change
        # this in the future. As another alternative, I could also cloudpickle
        # input objects when they are not JSON serializable.
        # OPTIMIZE: Prefect current tries to JSON serialize within their
        # client.create_flow_run method. In the future, we may want to move
        # this functionality there.

        parameters_serialized = {}
        for parameter_key, parameter_value in parameters.items():

            try:
                json.dumps(parameter_value)
            except TypeError:
                if hasattr(parameter_value, "as_dict"):
                    parameter_value = parameter_value.as_dict()
                elif hasattr(parameter_value, "to_dict"):
                    parameter_value = parameter_value.to_dict()
                elif parameter_key == "directory":
                    parameter_value = str(parameter_value)  # convert Path to str
                elif parameter_key == "source":
                    # recursive call to this function
                    parameter_value = cls._serialize_parameters(**parameter_value)

                # workflow_base and input_parameters are special cases that
                # may require a refactor (for customized workflows)
                elif parameter_key == "workflow_base":
                    parameter_value = parameter_value.name_full
                elif parameter_key == "input_parameters":
                    # recursive call to this function
                    parameter_value = cls._serialize_parameters(**parameter_value)

                else:
                    parameter_value = cloudpickle.dumps(parameter_value)
            parameters_serialized[parameter_key] = parameter_value
        return parameters_serialized

    @classmethod
    def _deserialize_parameters(
        cls,
        add_defaults_from_attr: bool = True,
        **parameters,
    ) -> dict:
        """
        converts all parameters to appropriate python objects
        """

        from simmate.toolkit import Structure
        from simmate.toolkit.diffusion import MigrationHop, MigrationImages

        parameters_cleaned = parameters.copy()

        #######
        # SPECIAL CASE: customized workflows have their parameters stored under
        # "input_parameters" instead of the base dict
        # THIS INVOLVES A RECURSIVE CALL TO THIS SAME METHOD
        if "workflow_base" in parameters.keys():
            # This is a non-modular import that can cause issues and slower
            # run times. We therefore import lazily.
            from simmate.workflows.utilities import get_workflow

            # Make sure we have a workflow object
            parameters_cleaned["workflow_base"] = (
                get_workflow(parameters["workflow_base"])
                if isinstance(parameters["workflow_base"], str)
                else parameters["workflow_base"]
            )
            # Make a recursive call for the input parameters
            parameters_cleaned["input_parameters"] = Workflow._deserialize_parameters(
                **parameters["input_parameters"]
            )
            return parameters_cleaned
        #######

        # For a series of parameters, we want their default values to be loaded
        # from class attributes if they are not set. To do this, we first check
        # if the parameter is set in our kwargs dictionary and check if the
        # value is set to "None". If it is, then we change to the value set as
        # the class attribute.
        if add_defaults_from_attr:
            for parameter in cls.parameter_names:
                if parameters.get(parameter, None) == None and hasattr(cls, parameter):
                    parameters_cleaned[parameter] = getattr(cls, parameter)

        # The remaining checks look to intialize input to toolkit objects

        structure = parameters.get("structure", None)
        if structure:
            parameters_cleaned["structure"] = Structure.from_dynamic(structure)
        else:
            parameters_cleaned.pop("structure", None)

        if "structures" in parameters.keys():
            structure_filenames = parameters["structures"].split(";")
            parameters_cleaned["structures"] = [
                Structure.from_dynamic(file) for file in structure_filenames
            ]

        if "migration_hop" in parameters.keys():
            migration_hop = MigrationHop.from_dynamic(parameters["migration_hop"])
            parameters_cleaned["migration_hop"] = migration_hop

        if "migration_images" in parameters.keys():
            migration_images = MigrationImages.from_dynamic(
                parameters["migration_images"]
            )
            parameters_cleaned["migration_images"] = migration_images

        if "supercell_start" in parameters.keys():
            parameters_cleaned["supercell_start"] = Structure.from_dynamic(
                parameters["supercell_start"]
            )

        if "supercell_end" in parameters.keys():
            parameters_cleaned["supercell_end"] = Structure.from_dynamic(
                parameters["supercell_end"]
            )

        if parameters.get("directory", None):
            parameters_cleaned["directory"] = Path(parameters_cleaned["directory"])

        if parameters.get("source", None):
            # !!! are there other types I should account for? Maybe I should just
            # make this a recursive call to catch everything?
            parameters_cleaned["source"]["directory"] = Path(
                parameters_cleaned["directory"]
            )

        return parameters_cleaned



# -*- coding: utf-8 -*-

"""
Workflows for analyzing ionic diffusion with a material
"""

from simmate.calculators.vasp.workflows.diffusion.all import (
    Diffusion__Vasp__NebAllPathsMit,
    Diffusion__Vasp__NebSinglePathMit,
    Diffusion__Vasp__NebFromEndpointsMit,
    Diffusion__Vasp__NebFromImagesMit,
)



# -*- coding: utf-8 -*-

"""
This is where you add any new database tables. This file is named
"models.py" because that what Django expects! In their terminology, a django 
'model' is the same thing as a 'database table'.

Whenever building new tables, be sure to start from the classes located in our
`simmate.database.base_data_types` module. These let you automatically add useful 
columns and features -- rather than creating everything from scratch. 

For more information and advanced guides, be sure to read through our
[base_data_types documentation](https://jacksund.github.io/simmate/simmate/database/base_data_types.html)
"""

from simmate.database.base_data_types import (
    DatabaseTable,
    table_column,
    Structure,
    Calculation,
)


# Mix-ins let you build out custom tables with common types of data. These will
# automatically build out your table with many different columns. Many
# also have a `from_toolkit` method that makes populating your database super easy
class MyCustomTable1(Structure, Calculation):
    pass  # nothing else is required unless you want to add custom columns/features


# alternatively, we can make a DatabaseTable from scratch
class MyCustomTable2(DatabaseTable):
    custom_column_01 = table_column.IntegerField()
    custom_column_02 = table_column.FloatField(null=True, blank=True)



# -*- coding: utf-8 -*-

from .matproj import Relaxation__Vasp__Matproj
from .matproj_hse import Relaxation__Vasp__MatprojHse
from .matproj_scan import Relaxation__Vasp__MatprojScan
from .matproj_metal import Relaxation__Vasp__MatprojMetal
from .mit import Relaxation__Vasp__Mit
from .mvl_grainboundary import Relaxation__Vasp__MvlGrainboundary
from .mvl_slab import Relaxation__Vasp__MvlSlab
from .mvl_neb_endpoint import Relaxation__Vasp__MvlNebEndpoint
from .quality00 import Relaxation__Vasp__Quality00
from .quality01 import Relaxation__Vasp__Quality01
from .quality02 import Relaxation__Vasp__Quality02
from .quality03 import Relaxation__Vasp__Quality03
from .quality04 import Relaxation__Vasp__Quality04
from .staged import Relaxation__Vasp__Staged



# -*- coding: utf-8 -*-

from .combine_chgcars import PopulationAnalysis__Bader__CombineChgcars
from .bader import PopulationAnalysis__Bader__Bader
from .badelf import PopulationAnalysis__Bader__Badelf



# -*- coding: utf-8 -*-

from simmate.utilities import get_doc_from_readme

__doc__ = get_doc_from_readme(__file__)

from simmate.database.base_data_types import (
    Relaxation,
    IonicStep,
    StaticEnergy,
    BandStructureCalc,
    DensityofStatesCalc,
    PopulationAnalysis,
    DynamicsRun,
    DynamicsIonicStep,
    DiffusionAnalysis,
    MigrationHop,
    MigrationImage,
    CustomizedCalculation,
)

from simmate.toolkit.structure_prediction.evolution.database import (
    EvolutionarySearch,
    StructureSource,
)



# -*- coding: utf-8 -*-

import time
import cloudpickle  # needed to serialize Prefect workflow runs and tasks
import logging

from django.db import transaction

from simmate.workflow_engine.execution.database import WorkItem

# This string is just something fancy to display in the console when a worker
# starts up.
# This uses "Small Slant" from https://patorjk.com/software/taag/
HEADER_ART = r"""
   _____                  __        _      __         __
  / __(_)_ _  __ _  ___ _/ /____   | | /| / /__  ____/ /_____ ____
 _\ \/ /  ' \/  ' \/ _ `/ __/ -_)  | |/ |/ / _ \/ __/  '_/ -_) __/
/___/_/_/_/_/_/_/_/\_,_/\__/\__/   |__/|__/\___/_/ /_/\_\\__/_/

"""


class SimmateWorker:

    # Ideally, this worker would involve multiple threads threads going. One
    # thread would update the queue database with a "heartbeat" to let it know
    # that it is still working on tasks. The other thread will run the given
    # workitems in serial. I could even allow for more threads so that this
    # worker can run multiple workitems at once and in parallel.
    # However, if this level of implementation is needed, we should instead
    # switch to using Prefect, which has it built in.

    def __init__(
        self,
        # limit of tasks and lifetime of the worker
        nitems_max: int = None,
        timeout: float = None,
        # wait_on_timeout=False, # TODO
        # settings on what to do when the queue is empty
        close_on_empty_queue: bool = False,
        waittime_on_empty_queue: float = 1,
        tags: list[str] = ["simmate"],  # should default be empty...?
    ):
        # the tags to query tasks for. If no tags were given, the worker will
        # query for tasks that have NO tags
        self.tags = tags

        # the maximum number of workitems to run before closing down
        # if no limit was set, we can go to infinity!
        self.nitems_max = nitems_max or float("inf")

        # Don't start a new workitem after this time. The worker will be shut down.
        # if no timeout was set, use infinity so we wait forever.
        self.timeout = timeout or float("inf")

        # whether to wait on the running workitems to finish before shutting down
        # the timedout worker.
        # self.wait_on_timeout = wait_on_timeout # # TODO

        # whether to close if the queue is empty
        self.close_on_empty_queue = close_on_empty_queue
        # if the queue is found to be empty, we will give it one last chance
        # to fill. Check the queue again after this time sleeping and if it is
        # still empty, close the worker.
        self.waittime_on_empty_queue = waittime_on_empty_queue

    def start(self):

        # print the header in the console to let the user know the worker started
        print(HEADER_ART)

        # loggin helpful info
        logging.info(f"Starting worker with tags {list(self.tags)}")

        # establish starting point for the worker
        time_start = time.time()
        ntasks_finished = 0

        # Loop endlessly until one of the following happens...
        #   the timeout limit is hit
        #   the queue is empty
        #   the nitems limit is hit
        while True:

            # check for timeout before starting a new workitem and exit
            # if we've hit the limit.
            if (time.time() - time_start) > self.timeout:
                # TODO - check wait_on_timeout if running in parallel.
                logging.info(
                    "The time-limit for this worker has been hit. Shutting down."
                )
                return

            # check the number of jobs completed so far, and exit if we hit
            # the limit
            if ntasks_finished >= self.nitems_max:
                logging.info(
                    f"Maximum number of WorkItems reached ({self.nitems_max}). "
                    "Shutting down."
                )
                return

            # check the length of the queue and while it is empty, we want to
            # loop. The exception of looping endlessly is if we want the worker
            # to shutdown instead.
            while self.queue_size() == 0:
                # if it is empty, we want to sleep for a little and check again
                time.sleep(self.waittime_on_empty_queue)

                # This is a special condition where we may want to close the
                # worker if the queue stays empty
                if self.close_on_empty_queue:
                    # after we just waited, let's check the queue size again
                    if self.queue_size() == 0:
                        # if it's still empty, we should close the worker
                        logging.info("The task queue is empty. Shutting down.")
                        return

            # If we've made it this far, we're ready to grab a new WorkItem
            # and run it!
            # Query for PENDING WorkItems, lock it for editting, and update
            # the status to RUNNING. And grab the first result
            workitem = (
                WorkItem.objects.select_for_update()
                .filter(status="P")
                .filter_by_tags(self.tags)
                .first()
            )

            # our lock exists only within this transation
            with transaction.atomic():
                # update the status to running before starting it so no other
                # worker tries to grab the same WorkItem
                workitem.status = "R"
                # TODO -- indicate that the WorkItem is with this Worker (relationship)
                workitem.save()

            # Print out the job ID that is being ran for the user to see
            logging.info(f"Running WorkItem with id {workitem.id}")

            # now let's unpickle the WorkItem components
            fxn = cloudpickle.loads(workitem.fxn)
            args = cloudpickle.loads(workitem.args)
            kwargs = cloudpickle.loads(workitem.kwargs)

            # Try running the WorkItem
            try:
                result = fxn(*args, **kwargs)
            # if it fails, we want to "capture" the error and return it
            # rather than have the Worker fail itself.
            except Exception as exception:
                result = exception

            # whatever the result, we need to try to pickle it now
            try:
                result_pickled = cloudpickle.dumps(result)
            # if this fails, we even want to pickle the error and return it
            except Exception as exception:
                result_pickled = cloudpickle.dumps(exception)

            # requery the WorkItem to restart our lock
            workitem = WorkItem.objects.select_for_update().get(pk=workitem.pk)

            # our lock exists only within this transation
            with transaction.atomic():
                # pickle the result and update the workitem's result and status
                # !!! should I have the pickle inside of a Try?
                workitem.result = result_pickled
                workitem.status = "F"
                workitem.save()

            # mark down that we've completed one WorkItem
            ntasks_finished += 1

            # Print out the job ID that was just finished for the user to see.
            logging.info("Completed WorkItem")

    def queue_size(self):
        """
        Return the approximate size of the queue.
        """
        # Count the number of WorkItem(s) that have a status of PENDING
        # !!! Should I include RUNNING in the count? If so I do that with...
        #   from django.db.models import Q
        #   ...filter(Q(status="P") | Q(status="R"))
        queue_size = (
            WorkItem.objects.filter(status="P").filter_by_tags(self.tags).count()
        )
        return queue_size

    @classmethod
    def run_singleflow_worker(cls):
        worker = cls(
            nitems_max=1,
            close_on_empty_queue=True,
            tags=["simmate"],
        )
        worker.start()



# -*- coding: utf-8 -*-

from simmate.conftest import copy_test_files, make_dummy_files

from simmate.toolkit.diffusion import MigrationImages

from simmate.calculators.vasp.inputs import Potcar
from simmate.calculators.vasp.workflows.diffusion.neb_from_images_mit import (
    Diffusion__Vasp__NebFromImagesMit,
)
from simmate.calculators.vasp.inputs.potcar_mappings import (
    PBE_ELEMENT_MAPPINGS_LOW_QUALITY,
)


def test_neb_setup(sample_structures, tmp_path, mocker):

    # To test this task we need to create images, which we do using I diffusion
    # in Y2CI2. We use [0] to grab the shortest path.
    structure = sample_structures["Y2CI2_mp-1206803_primitive"]
    images = MigrationImages.from_structure(structure, "I")[0]

    # estabilish filenames that we make and commonly reference
    incar_filename = tmp_path / "INCAR"
    potcar_filename = tmp_path / "POTCAR"
    # These files exist within a series of directories 00, 01,..., 05
    poscar_filenames = [tmp_path / str(n).zfill(2) / "POSCAR" for n in range(5)]

    # Because we won't have POTCARs accessible, we need to cover this function
    # call -- specifically have it pretend to make a file
    mocker.patch.object(
        Potcar,
        "to_file_from_type",
        return_value=make_dummy_files(potcar_filename),
    )

    # try to make input files in the tmp_path
    Diffusion__Vasp__NebFromImagesMit.setup(
        migration_images=images,
        directory=tmp_path,
    )
    assert incar_filename.exists()
    assert potcar_filename.exists()
    assert all([f.exists() for f in poscar_filenames])
    Potcar.to_file_from_type.assert_called_with(
        structure.composition.elements,
        "PBE",
        potcar_filename,
        PBE_ELEMENT_MAPPINGS_LOW_QUALITY,
    )


def test_neb_workup(tmp_path):
    copy_test_files(
        tmp_path,
        test_directory=__file__,
        test_folder="nudged_elastic_band.zip",
    )

    # estabilish filenames that we make and commonly reference
    summary_filename = tmp_path / "simmate_summary.yaml"
    plot_filename = tmp_path / "NEB_plot.jpeg"
    cif_filename = tmp_path / "path_relaxed_neb.cif"

    # run the full workup
    Diffusion__Vasp__NebFromImagesMit.workup(tmp_path)
    assert summary_filename.exists()
    # assert os.path.exists(plot_filename)  # temporarily removed
    assert cif_filename.exists()



# -*- coding: utf-8 -*-

import logging

from pymatgen.analysis.diffusion.neb.pathfinder import (
    MigrationHop as PymatgenMigrationHop,
)


class MigrationHop(PymatgenMigrationHop):
    def as_dict(self):
        # Pymatgen.diffusion has a broken as_dict method, so I need to fix this
        # here. I don't want to raise an error because I don't want issues
        # with the higher-level serialize_parameters utility
        return (
            "MigrationHop is not MSONable, and this bug is within pymatgen diffusion."
            "See https://github.com/materialsvirtuallab/pymatgen-analysis-diffusion/pull/264"
        )

    @classmethod
    def from_dynamic(cls, migration_hop):
        """
        This is an experimental feature. The code here is a repurposing of
        Structre.from_dynamic so consider making a general class for
        from_dynamic methods.
        """
        is_from_past_calc = False

        if isinstance(migration_hop, PymatgenMigrationHop):
            migration_hop_cleaned = migration_hop
        elif isinstance(migration_hop, dict) and "@module" in migration_hop.keys():
            migration_hop_cleaned = cls.from_dict(migration_hop)
        elif (
            isinstance(migration_hop, dict)
            and "migration_hop_table" in migration_hop.keys()
        ):
            is_from_past_calc = True
            migration_hop_cleaned = cls.from_database_dict(migration_hop)
        else:
            raise Exception("Unknown format provided for migration_hop input.")
        migration_hop_cleaned.is_from_past_calc = is_from_past_calc

        return migration_hop_cleaned

    @classmethod
    def from_database_dict(cls, migration_hop: dict):
        """
        This is an experimental feature. The code here is a repurposing of
        Structre.from_dynamic so consider making a general class for
        from_dynamic methods.

        ex:
            {
                "migration_hop_table": "MITMigrationHop",
                "migration_hop_id": 1,
            }
        """

        # Imports are done locally to keep this class modular.
        from simmate.database import connect
        from simmate.website.workflows import models as all_datatables
        from django.utils.module_loading import import_string

        datatable_str = migration_hop["migration_hop_table"]

        if hasattr(all_datatables, datatable_str):
            datatable = getattr(all_datatables, datatable_str)
        else:
            datatable = import_string(datatable_str)
        # for now I only support migration_hop_id
        migration_hop_db = datatable.objects.get(id=migration_hop["migration_hop_id"])
        migration_hop_cleaned = migration_hop_db.to_toolkit()
        migration_hop_cleaned.database_entry = migration_hop_db

        return migration_hop_cleaned

    def get_sc_structures(
        self,
        vac_mode: bool = True,
        min_atoms: int = 80,
        max_atoms: int = 240,
        min_length: float = 10.0,
        tol: float = 1e-5,
    ):

        supercell_start, supercell_end, supercell_base = super().get_sc_structures(
            vac_mode, min_atoms, max_atoms, min_length, tol
        )

        # The parent method doesn't work as itended... I don't want to fix it,
        # so I just warn users.
        if supercell_start.num_sites < min_atoms or supercell_end.num_sites > max_atoms:
            logging.warn(
                f"Failed to meet min_atoms={min_atoms} and max_atoms={max_atoms} "
                "requirements when making supercell. Resulted in"
                f" {supercell_start.num_sites} atoms."
            )
        if min(supercell_start.lattice.lengths) < min_length:
            logging.warn(
                "Failed to meet min_length={min_length} requirement when"
                " making supercell. Resulted in lattice vectors lengths of "
                f"{supercell_start.lattice.lengths}."
            )

        # BUG-CHECK: There's some rounding issue in the base code that I don't want
        # to mess with, so I just hack fix here.
        try:
            assert supercell_start != supercell_end
        except:
            raise Exception(
                "This structure has a bug due to a rounding error. "
                "Our team is aware of this bug and it has been fixed for the next "
                "pymatgen-analysis-diffusion release. In the meantime, we've found "
                "that increasing your supercell tolerances (min_atoms, max_atoms, min_length)"
                "can sometimes fix this error."
            )
        ####

        return supercell_start, supercell_end, supercell_base



# -*- coding: utf-8 -*-

# import pickle
import cloudpickle  # needed to serialize Prefect workflow runs and tasks

from simmate.database.base_data_types import DatabaseTable, table_column

# BUG: I have this database table within a module that calls "database.connect"
# at a higher level... Will this cause circular import issues?


class WorkItem(DatabaseTable):
    # The name WorkItem comes from the naming convention use here:
    # https://github.com/python/cpython/blob/master/Lib/concurrent/futures/thread.py

    # For serialization, I just use the pickle module, but in the future, I may
    # want to do a priority of MsgPk >> JSON >> Pickle. I could check if the given
    # object(s) has a serialize() method, check if has a to_json() method, and then
    # as a last resort just pickle the object.
    # https://docs.python.org/3/library/pickle.html

    # Pickled objects are just written as byte strings, so I stored them in django
    # as a BinaryField which accepts bytes.
    # https://docs.djangoproject.com/en/3.1/ref/models/fields/#binaryfield

    # There is a repo that goes one step further and makes a PickleField for django
    # but this looks over the top for what I need -- even though django-picklefield
    # is only one file! Still, I can use it as a reference if I ever run into bugs
    # https://github.com/gintas/django-picklefield

    class Meta:
        app_label = "workflow_engine"

    tags = table_column.JSONField(default=[])
    """
    List of tags to submit the task with, which helps with submitting workers
    for a specific type of task/workflow. (e.g. ["simmate", "custom"])
    """

    fxn = table_column.BinaryField()
    """
    The function to be called, which is serialized into a binary format.
    """

    args = table_column.BinaryField(default=cloudpickle.dumps([]))
    """
    positional arguments to be passed into fxn
    """

    kwargs = table_column.BinaryField(default=cloudpickle.dumps({}))
    """
    keyword arguments to be passed into fxn
    """

    result = table_column.BinaryField(blank=True, null=True)
    """
    the output of fxn(*args, **kwargs)
    """

    source = None
    """
    Source column is not needed so setting this to None disables the column
    """

    # These states are based on the python queue module
    class StatusOptions(table_column.TextChoices):
        PENDING = "P"
        RUNNING = "R"
        CANCELLED = "C"
        # CANCELLED_AND_NOTIFIED = "N"  # !!! when should I use this?
        FINISHED = "F"

    # TODO -- I should consider indexing this column for speed because it's
    # the most queried column by far.
    status = table_column.CharField(
        max_length=1,
        choices=StatusOptions.choices,
        default=StatusOptions.PENDING,
    )
    """
    the status/state of the workitem
    """

    # TODO: Consider creating a separate table for Worker and linking it to this
    # table via a foreign key.
    # worker_id = table_column.CharField(max_length=50, blank=True, null=True)



# -*- coding: utf-8 -*-

from simmate.conftest import copy_test_files, make_dummy_files
from simmate.calculators.vasp.inputs import Incar
from simmate.calculators.vasp.error_handlers import Eddrmm


def test_eddrmm(tmp_path):
    copy_test_files(
        tmp_path,
        test_directory=__file__,
        test_folder="eddrmm",
    )

    # we reference the files several spots below so we grab its path up front
    incar_filename = tmp_path / "INCAR"
    chgcar_filename = tmp_path / "CHGCAR"
    wavecar_filename = tmp_path / "WAVECAR"

    # init class with default settings
    error_handler = Eddrmm()

    # Confirm an error IS NOT found
    error_handler.filename_to_check = "vasp.no_error"
    assert error_handler.check(tmp_path) == False

    # Confirm an error IS found
    error_handler.filename_to_check = "vasp.out"
    assert error_handler.check(tmp_path) == True

    # Make first attempt at fixing the error
    make_dummy_files(chgcar_filename, wavecar_filename)
    fix = error_handler.correct(tmp_path)
    assert fix == "switched ALGO to Normal and deleted CHGCAR + WAVECAR"
    assert Incar.from_file(incar_filename)["ALGO"] == "Normal"
    assert not chgcar_filename.exists()
    assert not wavecar_filename.exists()

    # Make second attempt at fixing the error
    make_dummy_files(chgcar_filename, wavecar_filename)
    fix = error_handler.correct(tmp_path)
    assert fix == "switch POTIM from 0.5 to 0.25 and deleted CHGCAR + WAVECAR"
    assert Incar.from_file(incar_filename)["POTIM"] == 0.25
    assert not chgcar_filename.exists()
    assert not wavecar_filename.exists()


def test_eddrmm_neb(tmp_path):
    copy_test_files(
        tmp_path,
        test_directory=__file__,
        test_folder="eddrmm_neb",
    )
    # This test is identical to test_eddrmm but with an NEB folder organization

    # we reference the files several spots below so we grab its path up front
    incar_filename = tmp_path / "INCAR"
    # These files exist within a series of directories 00, 01,..., 05
    chgcar_filenames = [tmp_path / str(n).zfill(2) / "CHGCAR" for n in range(5)]
    wavecar_filenames = [tmp_path / str(n).zfill(2) / "WAVECAR" for n in range(5)]

    # init class with default settings
    error_handler = Eddrmm()

    # Confirm an error IS NOT found
    error_handler.filename_to_check = "vasp.no_error"
    assert error_handler.check(tmp_path) == False

    # Confirm an error IS found
    error_handler.filename_to_check = "vasp.out"
    assert error_handler.check(tmp_path) == True

    # Make first attempt at fixing the error
    make_dummy_files(*chgcar_filenames, *wavecar_filenames)
    fix = error_handler.correct(tmp_path)
    assert (
        fix == "switched ALGO to Normal and deleted CHGCARs + WAVECARs for all images"
    )
    assert Incar.from_file(incar_filename)["ALGO"] == "Normal"
    assert not any([f.exists() for f in chgcar_filenames])
    assert not any([f.exists() for f in wavecar_filenames])

    # Make second attempt at fixing the error
    make_dummy_files(*chgcar_filenames, *wavecar_filenames)
    fix = error_handler.correct(tmp_path)
    assert (
        fix
        == "switch POTIM from 0.5 to 0.25 and deleted CHGCARs + WAVECARs for all images"
    )
    assert Incar.from_file(incar_filename)["POTIM"] == 0.25
    assert not any([f.exists() for f in chgcar_filenames])
    assert not any([f.exists() for f in wavecar_filenames])



# -*- coding: utf-8 -*-


from simmate.database.base_data_types import StaticEnergy, table_column


class PopulationAnalysis(StaticEnergy):
    """
    This table combines results from a static energy calculation and the follow-up
    oxidation analysis on the charge density.
    """

    class Meta:
        app_label = "workflows"

    oxidation_states = table_column.JSONField(blank=True, null=True)
    """
    A list of calculated oxidation states based on some analysis. This is 
    given back as a list of float values in the same order as sites in the
    source structure.
    """



# -*- coding: utf-8 -*-

from pathlib import Path

from simmate.toolkit import Structure
from simmate.calculators.vasp.workflows.base import VaspWorkflow
from simmate.calculators.vasp.inputs import Incar, Poscar, Kpoints, Potcar


class DynamicsWorkflow(VaspWorkflow):
    @classmethod
    def setup(
        cls,
        structure: Structure,
        directory: Path,
        temperature_start: int = 300,
        temperature_end: int = 1200,
        time_step: float = 2,
        nsteps: int = 10000,
        **kwargs,
    ):

        # run cleaning and standardizing on structure (based on class attributes)
        structure_cleaned = cls._get_clean_structure(structure, **kwargs)

        # write the poscar file
        Poscar.to_file(structure_cleaned, directory / "POSCAR")

        # Combine our base incar settings with those of our parallelization settings
        # and then write the incar file. Note, we update the values of this incar,
        # so we make a copy of the dict.
        incar = cls.incar.copy()
        incar["TEBEG"] = temperature_start
        incar["TEEND"] = temperature_end
        incar["NSW"] = nsteps
        incar["POTIM"] = time_step
        incar = Incar(**incar) + Incar(**cls.incar_parallel_settings)
        incar.to_file(
            filename=directory / "INCAR",
            structure=structure,
        )

        # if KSPACING is not provided in the incar AND kpoints is attached to this
        # class instance, then we write the KPOINTS file
        if cls.kpoints and ("KSPACING" not in cls.incar):
            Kpoints.to_file(
                structure,
                cls.kpoints,
                directory / "KPOINTS",
            )

        # write the POTCAR file
        Potcar.to_file_from_type(
            structure.composition.elements,
            cls.functional,
            directory / "POTCAR",
            cls.potcar_mappings,
        )



# -*- coding: utf-8 -*-

import pytest

from simmate.conftest import copy_test_files, make_dummy_files
from simmate.calculators.vasp.inputs import Potcar
from simmate.calculators.vasp.workflows.base import VaspWorkflow
from simmate.calculators.vasp.inputs.potcar_mappings import PBE_ELEMENT_MAPPINGS


class Testing__Vasp__Dummy(VaspWorkflow):
    """
    A minimal example VaspWorkflow that is just for testing
    """

    use_database = False

    functional = "PBE"
    potcar_mappings = PBE_ELEMENT_MAPPINGS
    confirm_convergence = True
    pre_sanitize_structure = True
    incar = dict(
        PREC="Low",
        EDIFF__per_atom=2e-3,  # to ensure structure-specific kwargs
        KSPACING=0.75,
    )


# For shorthand reference below
DummyWorkflow = Testing__Vasp__Dummy


def test_base_setup(structure, tmp_path, mocker):

    # estabilish filenames that we make and commonly reference
    incar_filename = tmp_path / "INCAR"
    poscar_filename = tmp_path / "POSCAR"
    potcar_filename = tmp_path / "POTCAR"

    # Because we won't have POTCARs accessible, we need to cover this function
    # call -- specifically have it pretend to make a file
    mocker.patch.object(
        Potcar,
        "to_file_from_type",
        return_value=make_dummy_files(potcar_filename),
    )

    # try to make input files in the tmp_path
    DummyWorkflow.setup(directory=tmp_path, structure=structure)
    assert incar_filename.exists()
    assert poscar_filename.exists()
    assert potcar_filename.exists()
    Potcar.to_file_from_type.assert_called_with(
        structure.composition.elements,
        "PBE",
        potcar_filename,
        PBE_ELEMENT_MAPPINGS,
    )


def test_base_workup(tmp_path):
    copy_test_files(
        tmp_path,
        test_directory=__file__,
        test_folder="base.zip",
    )

    # estabilish filenames that we make and commonly reference
    summary_filename = tmp_path / "simmate_summary.yaml"
    vasprun_filename = tmp_path / "vasprun.xml"

    # run the full workup
    DummyWorkflow.workup(tmp_path)
    assert summary_filename.exists()

    # run the workup again with a malformed xml
    with vasprun_filename.open("r") as file:
        contents = file.readlines()
    with vasprun_filename.open("w") as file:
        file.writelines(contents[50])
    with pytest.raises(Exception):
        DummyWorkflow.workup(tmp_path)



# -*- coding: utf-8 -*-

import pytest

from simmate.command_line.workflow_engine import workflow_engine


@pytest.mark.django_db
def test_database_dump_and_load(command_line_runner):

    # dump the database to json
    result = command_line_runner.invoke(workflow_engine, ["start-singleflow-worker"])
    assert result.exit_code == 0

    # load the database to json
    result = command_line_runner.invoke(
        workflow_engine, ["start-worker", "-n", "1", "-e", "true"]
    )
    assert result.exit_code == 0



# -*- coding: utf-8 -*-

"""
This defines commands for managing your Simmate workflow engine. All commands are 
accessible through the "simmate workflow-engine" command.
"""

import click


@click.group()
def workflow_engine():
    """
    A group of commands for starting up Simmate Workers, Prefect Agents, and
    Dask Clusters. These are meant for setting up remote computational resources.
    """
    pass


@workflow_engine.command()
@click.option(
    "--nitems_max",
    "-n",
    default=None,
    type=int,
    help="the number of task run to submit before shutdown",
)
@click.option(
    "--timeout",
    "-t",
    default=None,
    type=float,
    help="the time (in seconds) after which this worker will stop running jobs and shutdown",
)
@click.option(
    "--close_on_empty_queue",
    "-e",
    default=False,
    type=bool,
    help="whether to shutdown when the queue is empty",
)
@click.option(
    "--waittime_on_empty_queue",
    "-w",
    default=1,
    type=float,
    help=(
        "if the queue is empty, the time (in seconds) the worker should wait"
        " before checking the queue again"
    ),
)
@click.option(
    "--tags",
    "-t",
    default=["simmate"],
    help="tags to filter tasks by for submission. defaults to just 'simmate'",
    multiple=True,
)
def start_worker(
    nitems_max,
    timeout,
    close_on_empty_queue,
    waittime_on_empty_queue,
    tags,
):
    """
    This starts a Simmate Worker which will query the database for jobs to run.
    """

    from simmate.workflow_engine import Worker

    worker = Worker(
        nitems_max,
        timeout,
        close_on_empty_queue,
        waittime_on_empty_queue,
        tags,
    )
    worker.start()


@workflow_engine.command()
def start_singleflow_worker():
    """
    This starts a Simmate Worker that only runs one job and then shuts down. Also,
    if no job is available in the queue, it will shut down.

    Note: this can be acheived using the start-worker command too, but because
    this is such a common use-case, we include this command for convienence.
    It is the same as...

    simmate workflow-engine start-worker -n 1 -e True -t simmate
    """

    from simmate.workflow_engine import Worker

    Worker.run_singleflow_worker()


# explicitly list functions so that pdoc doesn't skip them
__all__ = [
    "workflow_engine",
    "start_worker",
    "start_singleflow_worker",
]



# -*- coding: utf-8 -*-

from pathlib import Path

from simmate.toolkit import Structure
from simmate.workflow_engine import Workflow
from simmate.calculators.vasp.workflows.static_energy.matproj import (
    StaticEnergy__Vasp__Matproj,
)
from simmate.calculators.bader.workflows import (
    PopulationAnalysis__Bader__Bader,
    PopulationAnalysis__Bader__CombineChgcars,
)


class PopulationAnalysis__Vasp__BaderMatproj(Workflow):
    """
    Runs a static energy calculation using an extra-fine FFT grid and then
    carries out Bader analysis on the resulting charge density.
    """

    @classmethod
    def run_config(
        cls,
        structure: Structure,
        command: str = None,
        source: dict = None,
        directory: Path = None,
    ):

        prebader_result = PopulationAnalysis__Vasp__PrebaderMatproj.run(
            structure=structure,
            command=command,
            source=source,
            directory=directory,
        ).result()

        # Setup chargecars for the bader analysis and wait until complete
        PopulationAnalysis__Bader__CombineChgcars.run(
            directory=prebader_result["directory"],
        ).result()

        # Bader only adds files and doesn't overwrite any, so I just run it
        # in the original directory. I may switch to copying over to a new
        # directory in the future though.
        bader_result = PopulationAnalysis__Bader__Bader.run(
            directory=prebader_result["directory"]
        ).result()

        return bader_result

    @classmethod
    def _save_to_database(cls, bader_result, run_id):
        # load the results. We are particullary after the first result with
        # is a pandas dataframe of oxidation states.
        oxidation_data, extra_data = bader_result["result"]

        # load the calculation entry for this workflow run. This should already
        # exist thanks to the load_input_and_register task of the prebader workflow
        calculation = cls.database_table.from_run_context(
            run_id=run_id,
            workflow_name=cls.name_full,
        )
        # BUG: can't use context to grab the id because workflow tasks generate a
        # different id than the main workflow

        # now update the calculation entry with our results
        calculation.oxidation_states = list(oxidation_data.oxidation_state.values)
        calculation.save()


class PopulationAnalysis__Vasp__PrebaderMatproj(StaticEnergy__Vasp__Matproj):
    """
    Runs a static energy calculation with a high-density FFT grid under settings
    from the Materials Project. Results can be used for Bader analysis.

    We do NOT recommend running this calculation on its own. Instead, you should
    use the full workflow, which runs this calculation AND the following bader
    analysis for you. This S3Task is only the first step of that workflow.

    See `bader.workflows.materials_project`.
    """

    # The key thing for bader analysis is that we need a very fine FFT mesh. Other
    # than that, it's the same as a static energy calculation.
    incar = StaticEnergy__Vasp__Matproj.incar.copy()
    incar.update(
        NGXF__density_a=20,
        NGYF__density_b=20,
        NGZF__density_c=20,
        LAECHG=True,  # write core charge density to AECCAR0 and valence to AECCAR2
    )



# -*- coding: utf-8 -*-

from .neb_all_paths_mit import Diffusion__Vasp__NebAllPathsMit
from .neb_single_path_mit import Diffusion__Vasp__NebSinglePathMit
from .neb_from_endpoints_mit import Diffusion__Vasp__NebFromEndpointsMit
from .neb_from_images_mit import Diffusion__Vasp__NebFromImagesMit
from .neb_from_images_mvl_ci import Diffusion__Vasp__NebFromImagesMvlCi



# -*- coding: utf-8 -*-

from .matproj import StaticEnergy__Vasp__Matproj
from .matproj_hse import StaticEnergy__Vasp__MatprojHse
from .matproj_scan import StaticEnergy__Vasp__MatprojScan
from .mit import StaticEnergy__Vasp__Mit
from .quality04 import StaticEnergy__Vasp__Quality04
from .mvl_neb_endpoint import StaticEnergy__Vasp__MvlNebEndpoint



# -*- coding: utf-8 -*-

import shutil

from simmate.conftest import copy_test_files
from simmate.utilities.files import (
    get_directory,
    # make_archive,
    make_error_archive,
    archive_old_runs,
    empty_directory,
)


def test_get_directory(tmp_path):

    # create and delete a new directory
    new_directory = get_directory()
    assert "simmate-task-" in new_directory.name
    shutil.rmtree(new_directory)

    # get directory by name
    new_directory = get_directory(tmp_path)
    assert tmp_path == new_directory

    # test recursive creation
    subfolder = tmp_path / "subfolder1" / "subfolder2"
    new_directory = get_directory(subfolder)
    assert new_directory == subfolder


def test_make_archive(tmp_path):

    copy_test_files(
        tmp_path,
        test_directory=__file__,
        test_folder="to_archive",
    )

    archive_old_runs(tmp_path, time_cutoff=0)
    assert (tmp_path / "simmate-task-1.zip").exists()
    assert (tmp_path / "simmate-task-2.zip").exists()


def test_make_error_archive(tmp_path):

    copy_test_files(
        tmp_path,
        test_directory=__file__,
        test_folder="to_archive",
    )

    make_error_archive(tmp_path)
    assert (tmp_path / "simmate_attempt_01.zip").exists()

    make_error_archive(tmp_path)
    assert (tmp_path / "simmate_attempt_02.zip").exists()


def test_empty_directory(tmp_path):

    copy_test_files(
        tmp_path,
        test_directory=__file__,
        test_folder="to_archive",
    )

    empty_directory(tmp_path)
    assert not (tmp_path / "simmate-task-1").exists()
    assert not (tmp_path / "simmate-task-2").exists()
